(defpackage #:LED-cube
  (:use #:cl)
  (:nicknames :cube)
  (:export #:*cube-dimension* #:*current-cube-number* #:*current-intensity* #:*animation*
	   #:make-cube #:current-cube #:do-cube-times #:LED-number-to-coord #:sleep-for-delay #:goto-cube
	   #:with-disabled-save-undo #:undo
	   #:set-LED #:set-current-intensity
           #:copy-cube #:insert-cube-copy #:fill-cube #:empty-cube #:delete-cube #:delete-animation
	   #:move-cube
	   #:open-cube-device #:close-cube-device #:write-cube-to-device
           #:load-animation-file #:save-animation-file
           #:*transformations*))

(in-package :LED-cube)

(defvar *device-file* nil #+nil #p"/dev/ttyUSB0")
(defvar *device-lock* (clim-sys:make-lock "LED-cube device lock"))
(defvar *device-stream* nil)
(defvar *cube-dimension* 5)

(defvar *current-cube-number* 0)
(defvar *current-intensity* #\5)
(defvar *default-frame-delay* 0.09)

(defvar *undo-cache* nil)
(defvar *undo-depth* 30)
(defvar *save-undo-enabled* t)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; intensity functions

(defun ensure-intensity-to-be-of-character (intensity)
  (if (characterp intensity)
      intensity
      (code-char (+ intensity (char-code #\0)))))

(defun ensure-intensity-to-be-of-number (intensity)
  (if (numberp intensity)
      intensity
      (- (char-code intensity) (char-code #\0))))

(defun set-current-intensity (intensity)
  (setf *current-intensity*
        (ensure-intensity-to-be-of-character intensity)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; tools

(defun make-cube (&key (intensity #\0))
  (make-array (list *cube-dimension* *cube-dimension* *cube-dimension*)
	      :element-type 'character
	      :initial-element (ensure-intensity-to-be-of-character intensity)))

(defvar *animation* (list (make-cube)))
;;; We do perhaps need a lock for accesses on the *animation*, as the animation
;;; might be played by the idle-loop of the viewer while it gets modified inside
;;; the toolbox thread. (mgr, 20060102)

(defmacro current-cube ()
  '(nth *current-cube-number* *animation*))

(defmacro do-cube-times ((x y z &optional result) &rest body)
  `(dotimes (,z *cube-dimension* ,result)
      (dotimes (,y *cube-dimension*)
	(dotimes (,x *cube-dimension*)
	  ,@body))))

(defun LED-number-to-coord (number)
  (let* ((z (truncate (/ number (expt 5 2))))
	 (number (- number (* z (expt 5 2))))
	 (y (truncate (/ number 5)))
	 (number (- number (* y 5)))
	 (x number))
    (list x y z)))

(defun sleep-for-delay (&optional (delay *default-frame-delay*))
  (sleep delay))

(defun goto-cube (target)
  (cond
    ((or (eq :next target)
	 (eq :after target))
     (when (> (1- (length *animation*))  *current-cube-number*)
       (incf *current-cube-number*)))
    ((or (eq :previous target)
	 (eq :before target))
     (when (< 0  *current-cube-number*)
       (decf *current-cube-number*)))
    ((eq :last target)
     (setf *current-cube-number* (1- (length *animation*))))
    ((eq :first target)
     (setf *current-cube-number* 0))
    ((typep target 'number)
     (setf *current-cube-number* target))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; undo

(defmacro with-disabled-save-undo (&rest body)
  `(let ((*save-undo-enabled* nil))
     ,@body))

(defun save-undo ()
  (when *save-undo-enabled*
    (let (animation)
      (dolist (actual-cube *animation*)
	(pushnew (copy-cube actual-cube) animation))
      (pushnew (cons (nreverse animation) *current-cube-number*)
	       *undo-cache*))
    (when (> (length *undo-cache*) ;; limit *undo-cache to *undo-depth*
	     *undo-depth*)
      (setf *undo-cache* (subseq *undo-cache* 0 *undo-depth*)))))

(defun undo ()
  (when *undo-cache*
    (let ((actual-undo (pop *undo-cache*)))
      (setf *animation* (car actual-undo)
	    *current-cube-number* (cdr actual-undo)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; fundamental cube and animation alteration functions

(defun set-lED (x y z &optional (intensity *current-intensity*))
  (save-undo)
  (setf (aref (current-cube) x y z) (ensure-intensity-to-be-of-character intensity)))

(defun copy-cube (&optional (cube-to-copy (current-cube)))
  (let ((new-cube (make-cube)))
    (do-cube-times (x y z)
	(setf (aref new-cube x y z)
	      (aref cube-to-copy x y z)))
    new-cube))

(defun insert-cube-copy (&key (cube-to-copy (current-cube))
			      (position :after)
			      (move-to-new-cube t))
  (save-undo)
  (let (new-animation-list)
    (dolist (cube (reverse *animation*))
      (when (eq position :before)
	(pushnew cube new-animation-list))
      (when (eq cube (current-cube))
	(pushnew (copy-cube cube-to-copy) new-animation-list))
      (when (eq position :after)
	(pushnew cube new-animation-list)))
    (setf *animation* new-animation-list))
;  (setf *animation* (append *animation* (list (copy-cube))))
  (when move-to-new-cube
    (goto-cube position)))

(defmacro fill-cube (&key (intensity '*current-intensity*)
                          (cube-to-empty '(current-cube)))
  `(progn
     (save-undo)
     (setf ,cube-to-empty (make-cube :intensity ,intensity))))

(defmacro empty-cube (&optional (cube-to-empty '(current-cube)))
  `(fill-cube :cube-to-empty ,cube-to-empty
              :intensity #\0))

(defun delete-cube (&optional (number *current-cube-number*))
  (save-undo)
  (when (and (> (length *animation*) 1)
             (> (length *animation*) number)
             (<= 0 number))
    (multiple-value-prog1
        (values (setf *animation* (delete (nth number *animation*) *animation*))
                t)
      (when (<= number *current-cube-number*)
        (goto-cube :previous)))))

(defun delete-animation ()
  (save-undo)
  (setf *animation* (list (make-cube)))
  (goto-cube :first))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; additional cube and animation alteration functions
;; (This section shall be moved to a separate file.)

(defun move-cube (&key (cube-to-copy (current-cube)) (x 0) (y 0) (z 0))
  (save-undo)
  (let ((new-cube (make-cube)))
    (do-cube-times (x-m y-m z-m)
      (setf (aref new-cube (mod (+ z x-m) 5)
		           (mod (+ x y-m) 5)
			   (mod (+ y z-m) 5))
	    (aref cube-to-copy x-m y-m z-m)))
;    new-cube))
    (setf (current-cube) new-cube)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; input and output functions

(defun open-cube-device (&optional (device-file *device-file*))
  (close-cube-device)
  (setf *device-stream*
        (open device-file :direction :output :if-exists :append)))

(defun close-cube-device ()
  (when (and (streamp *device-stream*)
             (open-stream-p *device-stream*))
    (close *device-stream*))
  (setf *device-stream* nil))

(defun write-cube-to-device (&key (cube (current-cube))
			          (device-file *device-file*))
  (when device-file
    (clim-sys:with-lock-held (*device-lock*)
      (unless (and (streamp *device-stream*)
                   (open-stream-p *device-stream*))
        (open-cube-device device-file))
      (let ((serial *device-stream*))
;;      (with-open-file (serial device-file :direction :output :if-exists :append
;;                              ) ;; :element-type 'character)
        ;; hackhackhack
;;        (princ "0000" serial)
        (do-cube-times (x y z)
          ;; (princ (aref cube x y z))
          (princ (aref cube x y z) serial))
        ;; (princ #\Newline)
        (princ #\Newline serial)
        (force-output serial))
      cube))
    nil)

(defun load-animation-file (pathname)
  (let ((type (pathname-type pathname)))
    (unless (find type '("data" "raw") :test #'string-equal)
      (error (format nil "Don't know how to load a file of the type ~a." type)))

    (delete-animation)
    (with-disabled-save-undo
        (with-open-file (file (merge-pathnames pathname))
          (cond ((string-equal type "data")
                 (setf *animation* (read file)))
                
                ((string-equal type "raw")
                 (do ((line (read-line file) (read-line file nil 'eof)))
                     ((eq line 'eof))
                   ;; (print line)
                   (dotimes (led-number 125)
                     (destructuring-bind (x y z)
                         (led-number-to-coord led-number)
                       ;; (format t "-> ~d ~d ~d: ~d <-~%" x y z (elt line led-number)) (force-output)
                       (setf (aref (current-cube) x y z) (elt line led-number))))
                   (insert-cube-copy)))))
      (goto-cube :first))))

(defun save-animation-file (pathname &key (if-exists :error))
  (let ((type (pathname-type pathname)))
    (unless (find type '("data" "raw" "c") :test #'string-equal)
      (error (format nil "Don't know how to write a file of the type ~a." type)))

    (with-open-file (file (merge-pathnames pathname)
		        :direction :output :if-does-not-exist :create :if-exists if-exists)
    (cond ((string-equal type "data")
	   (write *animation* :stream file))

	  ((string-equal type "raw")
	   (dolist (cube *animation*)
	     (do-cube-times (x y z)
	       (princ (aref cube x y z) file))
	     (princ #\Newline file))
	   (force-output))

	  ((string-equal type "c")
	   (princ "{" file)
	   (dolist (cube *animation*)
	     (do-cube-times (x y z)
	       (princ (aref cube x y z) file)
	       (princ "," file))
	     (princ #\Newline file))
	   (princ "}" file)
	   (force-output))))))
